module Gallus
  # Public: The brain of logging. All defined loggers inherit from this class.
  class Log
    # Global context requires a semaphore to be thread-safe.
    @@global_context_mutex = Mutex.new

    # Global context stores context variables that will be printed by any logger alongside with its
    # local information.
    @@global_context = {}

    class << self
      # Public: Configures logger with given name. When no name specified, root logger will be configured.
      #
      # Example:
      #
      #   Gallus::Log.configure do |log|
      #     log.level = :INFO
      #     log.output << -> (event) { puts event.inspect }
      #   end
      #
      # Returns configured logger.
      def configure(name = '', &block)
        Repository.get_or_create_logger(name, &block)
      end

      # Public: Returns logger registered under given name.
      def [](name)
        configure(name)
      end

      # Internal: For testing purposes we should be able to remove logger with given name.
      def delete(name)
        Repository.delete_with_children(name)
      end

      # Internal: Performance magic. It dynamically defines log level methods to avoid overhead of not-logged
      # levels. For example if log level is WARN it'll define empty `trace`, `debug` and `info` methods and fully
      # working `warn` and `error`. Effectively log methods call `log` under the hood.
      def define_log_methods!(log_level)
        Level.each do |level|
          method_name = level.name.downcase
          remove_method(method_name) if method_defined?(method_name)

          if level >= log_level
            define_method(method_name) do |message = nil, payload = {}, &block|
              log(level, message, payload, &block)
            end
          else
            define_method(method_name) do |message = nil, payload = {}, &block|
              # supressed...
            end
          end
        end
      end

      # Public: Yields global context. Only block calls are allowed since this operation must be thread-safe.
      #
      # Example:
      #
      #   log.global_context { |ctx| ctx[:location] = "Castle Black" }
      #   samwell_log.info("I always wanted to be a Wizard", name: "Samwell Tarly")
      #   # => I always wanted to be a Wizard; name="Samwell Tarly" location="Castle Black"
      #   jon_snow_log.info("I know nothing", name: "Jon Snow")
      #   # => I know nothing; name="Jon Snow" location="Castle Black"
      #
      def global_context
        @@global_context_mutex.synchronize { yield @@global_context }
      end

      # Public: Yields context for current thread. Just to conform with the behavior of #global_context we're
      # using block call here as well.
      #
      # Example:
      #
      #   log.current_thread_context { |ctx| ctx[:location] = "Castle Black" }
      #   log.info("I know nothing!", name: "Jon Snow")
      #   # => I know nothing! name="Jon Snow" location="Castle Black"
      #
      def current_thread_context
        yield (Thread.current[:log_context] ||= {})
      end

      # Public: Returns root logger.
      def root
        @@root
      end
    end

    # Public: Logger name, log level and context serializtion handler.
    attr_accessor :name, :level

    # Public: A list with configured outputs.
    attr_reader :output

    # Internal: Parent logger. Used for testing purposes.
    attr_reader :parent

    # Internal: Constructs new logger with configuration inherited from given parent. Yields itself so
    # it can be reconfigured with a block. Don't use this method directly. Use Gallus::Log.configure class
    # method if you wanna define or reconfigure a logger.
    def initialize(parent, name, &block)
      @parent, @name = parent, name.to_s

      if parent
        @output = @parent.output.dup
        self.level = @parent.level
      end

      @output ||= []

      yield self if block_given?
    end

    # Public: Changing log level means that we have to redefine logging methods.
    def level=(level)
      level = level.is_a?(Level) ? level : Level[level.to_s]
      return if @level == level
      self.class.class_eval { define_log_methods!(level) }
      @level = level
    end

    # Public: Shortand to Gallus::Log.global_context.
    def global_context(&block)
      self.class.global_context(&block)
    end

    # Public: Shortand to Gallus::Log.current_thread_context.
    def current_thread_context(&block)
      self.class.current_thread_context(&block)
    end

    private

    # Internal: Merges data from all contexts and provides them as a single hash. Order:
    #
    # 1. Global context.
    # 2. Curreant thread context.
    # 3. Self context.
    #
    def merged_context
      {}.tap do |merged|
        self.class.global_context { |ctx| merged.merge!(ctx) }
        self.class.current_thread_context { |ctx| merged.merge!(ctx) }
      end
    end

    # Internal: Actual logging. Takes payload and merges it into contexts, serializes context variables
    # using configured serialization handler and finally sends it to each defined output.
    def log(level, message = nil, payload = {}, &block)
      message ||= block.call if block_given?
      event = Event.new(@name, level, message.to_s, merged_context.merge(payload))
      output.each { |out| out.call(event) }
    end

    # Root logger configuration (defaults for any logger defined after):

    @@root = Log.configure do |log|
      log.name = 'root'
      log.level = Level::INFO
    end
  end
end
