module Gallus
  # Internal: Wee need a place to keep track of registered loggers and their parents. This little
  # repository class handles this task in a thread-safe manner.
  class Repository
    PARENT_DELIMITER = '::'

    @mutex = Mutex.new

    # Internal: Returns all registered loggers.
    def self.all
      @all ||= {}
    end

    # Internal: There must be a way to find parent logger for given class name. For example, looking up parent
    # for Foo::Bar::Baz logger it'll look up for Foo::Bar, then falling back to Foo and eventually to root logger
    # if nothing found.
    def self.find_parent(name)
      parent, name = nil, name.dup

      while parent.nil?
        name = name.split(PARENT_DELIMITER)[0..-2].join(PARENT_DELIMITER)
        parent = all[name]
        break if name.empty?
      end

      parent
    end

    # Internal: For testing purposes we need to be able to delete logger alongside with all its children.
    def self.delete_with_children(name)
      @mutex.synchronize do
        all.keys.each { |k| all.delete(k) if k.start_with?(name + PARENT_DELIMITER) }
        all.delete(name)
      end
    end

    # Internal: We obviously need a way to create logger or retrieve it by name if already registered.
    # Creation of logger causes configuration to be inherited from parent (or root) logger.
    def self.get_or_create_logger(name, &block)
      name, log = name.to_s, nil

      @mutex.synchronize do
        if log = all[name]
          yield log if block_given?
        else
          all[name] = (log = Class.new(Log).new(find_parent(name), name, &block))
        end
      end

      log
    end
  end
end
